from decimal import Decimal

from eth_account.messages import SignableMessage, _hash_eip191_message, encode_typed_data

from jam_sdk.constants import (
    JAM_BALANCE_MANAGER,
    JAM_ORDER_TYPES,
    JAM_SETTLEMENT_CONTRACT,
    PERMIT2_ADDRESS,
    PERMIT2_WITH_JAM_ORDER_TYPES,
)


def apply_slippage(amount: int, slippage: Decimal) -> int:
    min_amount = int(Decimal(amount) * (Decimal(1) - (slippage / Decimal(100))))
    return min_amount


def hash_jam_order(chain_id: int, order: dict) -> str:
    jam_settlement_domain: dict = {
        "name": "JamSettlement",
        "version": "2",
        "chainId": chain_id,
        "verifyingContract": JAM_SETTLEMENT_CONTRACT[chain_id],
    }
    structured_data: SignableMessage = encode_typed_data(jam_settlement_domain, JAM_ORDER_TYPES, order)
    return _hash_eip191_message(structured_data).hex()


def hash_permit2_with_witness(chain_id: int, permit2_with_order_witness: dict) -> str:
    jam_settlement_domain: dict = {
        "name": "Permit2",
        "chainId": chain_id,
        "verifyingContract": PERMIT2_ADDRESS[chain_id],
    }
    structured_data: SignableMessage = encode_typed_data(
        jam_settlement_domain, PERMIT2_WITH_JAM_ORDER_TYPES, permit2_with_order_witness
    )
    return _hash_eip191_message(structured_data).hex()


def hash_permit2_jam_order(chain_id: int, order: dict) -> str:
    permit2_with_order_witness = {
        "permitted": [
            {"token": token, "amount": amount}
            for token, amount in zip(order["sellTokens"], order["sellAmounts"], strict=True)
        ],
        "spender": JAM_BALANCE_MANAGER[chain_id],
        "nonce": order["nonce"],
        "deadline": order["expiry"],
        "witness": order,
    }
    return hash_permit2_with_witness(chain_id, permit2_with_order_witness)


def hash_tosign_field(chain_id: int, to_sign: dict) -> str:
    if "witness" in to_sign:
        return hash_permit2_with_witness(chain_id, to_sign)
    return hash_jam_order(chain_id, to_sign)
